"""
 Vehicle Routing Problem
"""
from algorithm.core.Location import Tree
from algorithm.core.Vehicle import Vehicle
from algorithm.exceptions.CustomExceptions import DuplicateVehicle, DuplicateLocation, DuplicateShipment


class VehicleRoutingProblem:
    def __init__(self):
        self.vehicles = []
        self.locations = []
        self.shipments = []
        self.tree = Tree()

    def add_vehicle(self, new_vehicle):
        if new_vehicle.id not in [x.id for x in self.vehicles]:
            self.vehicles.append(new_vehicle)
            return "added"
        else:
            raise DuplicateVehicle("Vehicle with ID " + str(new_vehicle.id) + " was already inserted")

    def add_location(self, new_location):
        if new_location.id not in [x.id for x in self.locations]:
            self.locations.append(new_location)
            return "added"
        else:
            raise DuplicateLocation("Location with ID " + str(new_location.id) + " was already inserted")

    def add_shipment(self, new_shipment):
        if new_shipment.id not in [x.id for x in self.shipments]:
            self.shipments.append(new_shipment)
            return "added"
        else:
            raise DuplicateShipment("Shipment with ID " + str(new_shipment.id) + " was already inserted")

    def solve(self):
        self.__create_tree()

        path = self.tree.generate_whole_path()
        deport = self.tree.find_node(0)
        index = 0
        for vehicle in self.vehicles:
            vehicle.add_shipment(deport)
            next_location = self.tree.find_node(path[index])
            status, cap = vehicle.add_shipment(next_location)
            while status:
                if next_location.decrease_capacity(cap) == 0:

                    if path[index] == 0:
                        break
                    index += 1
                else:
                    break

                next_location = self.tree.find_node(path[index])
                status, cap = vehicle.add_shipment(next_location)

            node_id = vehicle.route[len(vehicle.route) - 1].id
            if node_id != 0:  # need to find path to deport
                while node_id != 0:
                    node = self.tree.find_node(node_id)
                    node_id = node.parent_id
                    vehicle.add_location(node_id)

            print(" ".join(map(str, vehicle.route)), vehicle.capacity, "|", vehicle.max_capacity)

        return path

    def solve2(self):
        self.__create_tree()

        path = self.tree.generate_whole_path(type = 2)
        print(path)
        deport = self.tree.find_node(0)
        index = 0
        for vehicle in self.vehicles:
            vehicle.add_shipment(deport)
            next_location = self.tree.find_node(path[index])
            status, cap = vehicle.add_shipment(next_location)
            while status:
                index += 1
                if index >= len(path):
                    break

                next_location = self.tree.find_node(path[index])
                status, cap = vehicle.add_shipment(next_location)

            node_id = vehicle.route[len(vehicle.route) - 1].id
            if node_id != 0:  # need to find path to deport
                while node_id != 0:
                    node = self.tree.find_node(node_id)
                    node_id = node.parent_id
                    vehicle.add_location(node_id)
            if index >= len(path):
                break
            print(" ".join(map(str, vehicle.route)), vehicle.capacity, "|", vehicle.max_capacity)

        return path

    def solve3(self):
        self.__create_tree()
        self.tree.root.sum_capacity = self.tree.calculate_capacity(self.tree.root)
        self.dfs(self.tree.root, 0)
        for vehicle in self.vehicles:
            print(" ".join(map(str, vehicle.route)), vehicle.capacity, "|", vehicle.max_capacity)

    def solve4(self):
        self.__create_tree()

        print(self.dfs2(self.tree.root))

    def dfs(self, node, index):
        for child in sorted(node.children, key=lambda x: x.id):

            if self.vehicles[index].capacity == 0:
                node.sum_capacity -= self.dfs(child, index)
            else:
                if node.capacity != 0:
                    status, cap = self.vehicles[index].add_shipment(node)
                    if cap != node.capacity:
                        status = False

                    while not status:
                        index += 1
                        status, cap = self.vehicles[index].add_shipment(node)
                        if cap != node.capacity:
                            status = False
                else:
                    if self.vehicles[index].max_capacity - self.vehicles[index].capacity >= child.sum_capacity:
                        node.sum_capacity -= self.dfs(child, index)
                    else:
                        index += 1
                        node.sum_capacity -= self.dfs(child, index)

        status, cap = self.vehicles[index].add_shipment(node)
        if cap != node.capacity:
            status = False

        while not status:
            index += 1
            status, cap = self.vehicles[index].add_shipment(node)
            if cap != node.capacity:
                status = False
        if len(node.children) == 0:
            return node.capacity
        else:
            return node.sum_capacity

    def dfs2(self, node):
        routes = []
        for child in sorted(node.children, key=lambda x: x.capacity):
            routes.append(self.dfs2(child))
        routes.append((node.id, node.capacity))
        if len(node.children) == 0:
            return [(node.id, node.capacity)]
        else:
            return routes






    def __create_tree(self):

        locations_to_add = self.locations.copy()

        deport = list(filter(lambda x: x.id == 0, locations_to_add))[0]
        self.tree.add_node(0, 0, None)
        locations_to_add.remove(deport)

        while len(locations_to_add):

            for x in filter(lambda y: y.parent_id in self.tree.ids, locations_to_add):
                self.tree.add_node(x.capacity, x.id, x.parent_id)
                locations_to_add.remove(x)




